#fc.py

import os
import re
import sys
import collections
import argparse
import tables
import itertools
import numba
import glob

import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.sparse as sp_sparse

from collections import defaultdict
from scipy import sparse, io
from scipy.sparse import csr_matrix
import matplotlib.pyplot as plt

from pySpade.utils import get_logger, read_sgrna_dict, find_single_sgrna_cells, find_all_sgrna_cells, set_axis_style, sc_exp_box_plot

logger = get_logger(logger_name=__name__)

def Calculate_fc(TRANSCRIPTOME_DIR, 
                 SGRNA_DICT, 
                 TARGET_FILE, 
                 OUTPUT_FOLDER):

    TRANSCRIPTOME_FILE = glob.glob(TRANSCRIPTOME_DIR + 'Singlet_sub_df.*')
    if (TRANSCRIPTOME_FILE[0].endswith('pkl')) == True:
        sub_df = pd.read_pickle(TRANSCRIPTOME_FILE[0])
    elif (TRANSCRIPTOME_FILE[0].endswith('h5')) == True:
        sub_df = pd.read_hdf(TRANSCRIPTOME_FILE[0], 'df')

    total_seq_reads = sub_df.sum(axis=0)
    SGRNA_FILE = glob.glob(TRANSCRIPTOME_DIR + 'Singlet_sgRNA*')
    if (SGRNA_FILE[0].endswith('pkl')):
        sgrna_df_bool = pd.read_pickle(SGRNA_FILE[0]) > 0 
    elif (SGRNA_FILE[0].endswith('h5')):
        sgrna_df_bool = pd.read_hdf(SGRNA_FILE[0], 'df') > 0

    if list(sub_df.columns) != list(sgrna_df_bool.columns):
        logger.critical('File format error, make sure transcriptome df columns are the same as sgrna df.')

    
    gene_seq = sub_df.index
    if len(gene_seq) != len(set(gene_seq)):
        logger.critical('Duplication of mapping genes.')
    unique_elements, counts = np.unique(gene_seq, return_counts=True)
    duplicate_elements = unique_elements[counts > 1]

    #load sgRNA dictionary
    sgrna_dict  = read_sgrna_dict(SGRNA_DICT)
    
    #load regions and their target genes
    query_region_list = []
    query_gene_list = []
    with open(TARGET_FILE) as ft:
        for line in ft:
            region, gene = line.strip().split('\t')
            if gene in duplicate_elements:
                logger.critical(str(gene) + ' duplication in the transcriptome. Skip calculating fold change.')
                continue
            query_region_list.append(region)
            query_gene_list.append(gene)

    logger.info('Finished loading data.')

    #prepare cpm array for each query gene
    All_cpm_array = []
    for gene in query_gene_list:
        cpm_array = sub_df.loc[gene ,:] / total_seq_reads * 1000000
        All_cpm_array.append(cpm_array)

    #calculate repression efficiency for single sgRNA
    All_FC_list = []
    All_sgRNA = []
    All_num_cell = []
    All_pert_cpm_list = []
    All_bg_cpm_list = []
    region_fc_list = []
    region_cell_list = []
    region_pval_list = []
    region_pert_cpm_list = []
    region_bg_cpm_list = []
    All_pval_list = []
    for i , target_gene, non_zero_array in zip(query_region_list, query_gene_list, All_cpm_array):
        logger.info(f'Calculate gene: {target_gene} in region: {i}')
        
        #Find if any sgRNA is missing from the sgRNA df
        missing_sgrna_list = list(set(sgrna_dict[i]) - set(sgrna_df_bool.index))
        if len(missing_sgrna_list) == len(sgrna_dict[i]):
            logger.info('Missing all the sgRNA in this region.')
            continue
        if len(missing_sgrna_list) > 0:
            logger.info('Missing ' + str(len(missing_sgrna_list)) + ' sgrna.')
            for j in missing_sgrna_list:
                logger.info('Missing sgrna: ' + str(j))

        #remove the missing sgrna and continue DE analysis 
        if len(missing_sgrna_list) > 0:
            remain_sgrna_list = list(set(sgrna_dict[i]) - (set(missing_sgrna_list)))
            sgrna_dict.update({i:remain_sgrna_list})

        sgrna_idx = find_all_sgrna_cells(sgrna_df_bool, sgrna_dict[i])
        other_idx = np.setxor1d(sgrna_idx, range(len(non_zero_array)))
        region_cell_num = len(sgrna_idx)
        region_cell_list.append(region_cell_num)

        sgrna_cells_expression = non_zero_array[sgrna_idx]
        other_cells_expression = non_zero_array[other_idx]
        fc = (np.mean(sgrna_cells_expression) + 0.01) / (np.mean(other_cells_expression) + 0.01)
        pval = stats.ttest_ind(sgrna_cells_expression, other_cells_expression, equal_var=False).pvalue
        region_fc_list.append(fc)
        region_pval_list.append(pval)
        region_pert_cpm_list.append(str(np.mean(sgrna_cells_expression)))
        region_bg_cpm_list.append(str(np.mean(other_cells_expression)))

        #Box plot for each region and the query gene 
        save_name = i.replace(':', '_')
        box_plot_file = OUTPUT_FOLDER + 'Box_plot-' + save_name + '-' + target_gene + '-10perc.tiff'
        sc_exp_box_plot(sgrna_cells_expression, other_cells_expression, target_gene, i, box_plot_file)

        sgrna_seq = []
        fc_list = []
        num_cell_list = []
        pert_cpm_list = []
        bg_cpm_list = []
        pval_list = []
        #Now we calculate fold change for individual sgrna in the perturbation
        for sgrna in sgrna_dict[i]:
            try:
                sgrna_idx = find_single_sgrna_cells(sgrna_df_bool, sgrna)
                unperturb_idx = np.setxor1d(sgrna_idx, range(len(non_zero_array)))
                perturb_cpm = np.mean(non_zero_array[sgrna_idx]) 
                bg_cpm = np.mean(non_zero_array[unperturb_idx])
                fc = (np.mean(non_zero_array[sgrna_idx]) + 0.01) / (np.mean(non_zero_array[unperturb_idx]) + 0.01)
                pval = stats.ttest_ind(non_zero_array[sgrna_idx], non_zero_array[unperturb_idx], equal_var=False).pvalue
                num_cell = str(len(sgrna_idx))
            except:
                fc = 1
                num_cell = '0'

            pert_cpm_list.append(perturb_cpm) 
            bg_cpm_list.append(bg_cpm)   
            fc_list.append(fc)
            pval_list.append(pval)
            sgrna_seq.append(sgrna)
            num_cell_list.append(num_cell)

        All_FC_list.append(fc_list)
        All_sgRNA.append(sgrna_seq)
        All_num_cell.append(num_cell_list)
        All_pval_list.append(pval_list)
        All_pert_cpm_list.append(pert_cpm_list)
        All_bg_cpm_list.append(bg_cpm_list)

    #write to output file
    output_file=open(OUTPUT_FOLDER + 'fold_change.txt', 'w')
    output_file.write('Region' + '\t' + 'sgRNA sequence' + '\t' + 'Target gene' + '\t' + 'Number of cells' + '\t' + 'Fold change' + '\t' + 'log(pval), t-test' + '\t' + 'Perturb cpm' + '\t' + 'Background cpm' + '\n')

    for idx in np.arange(len(query_region_list)):
        i = query_region_list[idx]
        target_gene = query_gene_list[idx]
        output_file.write(str(i) + '\t' + 'All_sgRNA_combine' + '\t' + str(target_gene) + '\t' + str(region_cell_list[idx]) + '\t' + str(region_fc_list[idx]) + '\t' + str(region_pval_list[idx]) + '\t' + str(region_pert_cpm_list[idx]) + '\t' + str(region_bg_cpm_list[idx])+ '\n')
        for sg in All_sgRNA[idx]:
            sec_idx = list(All_sgRNA[idx]).index(sg)
            num_cell = All_num_cell[idx][sec_idx]
            repression = All_FC_list[idx][sec_idx]
            pvalue = All_pval_list[idx][sec_idx]
            pert_cpm = All_pert_cpm_list[idx][sec_idx]
            sgbg_cpm = All_bg_cpm_list[idx][sec_idx]
            output_file.write(str(i) + '\t' + str(sg) + '\t' + str(target_gene)+ '\t' + str(num_cell) + '\t' + str(repression) + '\t' + str(pvalue) + '\t' + str(pert_cpm) + '\t' + str(sgbg_cpm) + '\n')
    
    output_file.close()

if __name__ == '__main__':
    pass



#DEobs.py

import os
import sys
import re
import collections
import argparse
import tables
import itertools
import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.sparse as sp_sparse
from multiprocessing import Pool
from collections import defaultdict
from scipy import sparse, io
from scipy.sparse import csr_matrix
from random import shuffle

from pySpade.utils import get_logger, read_annot_df, cpm_normalization, metacelll_normalization, read_sgrna_dict, find_all_sgrna_cells, perform_DE, hypergeo_test

np.random.seed(0)
logger = get_logger(logger_name=__name__)

def DE_observe_bootstrap_cells(sub_df_file,
                               sgrna_df,
                               sgrnas_file,
                               output_dir,
                               num_processing=1,
                               iteration=100,
                               rand_cell_num=500,
                               norm='cpm',
                               ):

    #check the normalization method   
    if (norm != 'cpm') and (norm != 'metacell'):
        logger.critical("Incorrect normalization method. Has to be either 'cpm' or 'metacell'")
        sys.exit(0)

    #read the 10X hdf5 file
    logger.info('Reading transcriptome file')

    if (sub_df_file.endswith('pkl')):
        sub_df = pd.read_pickle(sub_df_file)
    elif (sub_df_file.endswith('h5')):
        sub_df = pd.read_hdf(sub_df_file, 'df')
       
    idx = np.arange(0, len(sub_df.index))
    

    #normalize the matrix.       
    if (norm == 'cpm'):
        cpm_matrix = cpm_normalization(sub_df)
        
    elif (norm == 'metacell'):
        cpm_matrix = metacelll_normalization(sub_df)
     
    logger.info('Finished transcriptome normalization.')
    

    #load the sgRNA file
    logger.info('Loading sgRNA df.')
    if (sgrna_df.endswith('pkl')):
        sgrna_df_adj_bool = pd.read_pickle(sgrna_df) > 0 
    elif (sgrna_df.endswith('h5')):
        sgrna_df_adj_bool = pd.read_hdf(sgrna_df, 'df') > 0

    [g,c] = sub_df.shape
    assert np.sum(sub_df.columns == sgrna_df_adj_bool.columns) == c #make sure the sequences are the same compare with two df
    
    gene_num, cell_num = sub_df.shape
    del sub_df
    #perform hypergeometric test for every single gene in the dataframe
    sgrna_dict  = read_sgrna_dict(sgrnas_file)
    for k in sgrna_dict:
        logger.info(f'Start analysis of region: {k}')

        #Find if any sgRNA is missing from the sgRNA df
        missing_sgrna_list = list(set(sgrna_dict[k]) - set(sgrna_df_adj_bool.index))
        if len(missing_sgrna_list) == len(sgrna_dict[k]):
            logger.info('Missing all the sgRNA in this region.')
            continue
        if len(missing_sgrna_list) > 0:
            logger.info('Missing ' + str(len(missing_sgrna_list)) + ' sgrna.')
            for i in missing_sgrna_list:
                logger.info('Missing sgrna: ' + str(i))

        #remove the missing sgrna and continue DE analysis 
        if len(missing_sgrna_list) > 0:
            remain_sgrna_list = list(set(sgrna_dict[k]) - (set(missing_sgrna_list)))
            sgrna_dict.update({k:remain_sgrna_list})

        #idx index of cells containing the given sgRNA
        sgrna_idx = find_all_sgrna_cells(sgrna_df_adj_bool, sgrna_dict[k])

        combined_up_array = np.empty([iteration, gene_num])
        combined_down_array = np.empty([iteration, gene_num])
        combined_cpm_array = np.empty([iteration, gene_num])

        #Calculation of Bootstrapping
        logger.info(f'Start bootstrapping analysis of region: {k}')
        for i in np.arange(iteration):
            counter = i
            new_sgrna_idx = np.random.choice(sgrna_idx, size=rand_cell_num, replace=False)
            #force the up-tail p-vals of all zero expressed genes to be zero. (actual p-val is 1)
            pval_list_down = np.zeros(g)
            pval_list_up = np.zeros(g)
            fc_list = np.ones(g)
            cpm_list = np.zeros(g)
                
            #perform the differential gene analysis by using Virtual FACS
            num_sgrna_cell, pval_list_up, pval_list_down, fc_list, cpm_list = perform_DE(
                                                                                new_sgrna_idx,
                                                                                cpm_matrix,
                                                                                idx,
                                                                                num_processing,
                                                                                pval_list_down,
                                                                                pval_list_up,
                                                                                fc_list,
                                                                                cpm_list
                                                                            )
            #force the up-tail p-vals of all zero expressed genes to be zero. (actual p-val is 1)
            combined_up_array[(counter)] = pval_list_up
            combined_down_array[(counter)] = pval_list_down
            combined_cpm_array[(counter)] = cpm_list
            up_pval_matrix   = csr_matrix(combined_up_array)
            down_pval_matrix = csr_matrix(combined_down_array)
            region_cpm_matrix = csr_matrix(combined_cpm_array)

            #save all the output
            io.savemat(
                '%s/%s-bootstrapping-up_log-pval'%(output_dir, k[0:]),
                {'matrix':up_pval_matrix})

            io.savemat(
                '%s/%s-bootstrapping-down_log-pval'%(output_dir, k[0:]),
                {'matrix':down_pval_matrix})

            io.savemat(
                '%s/%s-bootstrapping-cpm'%(output_dir, k[0:]),
                {'matrix':region_cpm_matrix})

   
if __name__ == '__main__':
    pass



#__main__.py

import sys
import importlib
from pySpade import __version__
from pySpade.parsers import parse_args
from pySpade.utils import get_logger

def main():
    args = parse_args()

    logger = get_logger(logger_name=__name__)
    logger.info(f'pySpade version: {__version__}')
    logger.info(f'Python version: {sys.version_info.major}.{sys.version_info.minor}')

    if not sys.version_info.major == 3 and sys.version_info.minor >= 7:
        logger.critical('Please use Python >= 3.7')
        sys.exit(1)


    if (args.command == 'process'):
        logger.info('Running process command ...')
        m = importlib.import_module(name=f'pySpade.{args.command}')

        _ = m.process_singlet(
            WORK_DIR=args.feature_bc,
            SGRNA=args.input_sgrna,
            CELL_MULTIPLEX=args.cell_multiplex,
            OUTPUT_DIR=args.output_directory,
            COMP=args.comp
        )
        logger.info('Done.')


    elif (args.command == 'explevel'):
        logger.info('Running explevel command ...')
        m = importlib.import_module(name=f'pySpade.{args.command}')

        _ = m.Calculate_gene_expression(
            TRANSCRIPTOME_DIR=args.transcriptome_dir,
            GENE_FILE=args.gene,
            OUTPUT_FILE=args.output_file
        )
        logger.info('Done.')


    elif (args.command == 'fc'):
        logger.info('Running fc command ...')
        m = importlib.import_module(name=f'pySpade.{args.command}')

        _ = m.Calculate_fc(
            TRANSCRIPTOME_DIR=args.transcriptome_dir,
            SGRNA_DICT = args.dict,
            TARGET_FILE = args.region,
            OUTPUT_FOLDER = args.output_folder
        )        
        logger.info('Done.')


    elif (args.command == 'DEobs'):
        logger.info('Running DEobs command ...')
        m = importlib.import_module(name=f'pySpade.{args.command}')

        _ = m.DE_observe_cells(
            sub_df_file = args.transcriptome_df,
            sgrna_df = args.input_sgrna, 
            sgrnas_file = args.dict,
            num_processing = args.threads,
            norm = args.norm_method,
            output_dir = args.output_dir
            )        
        logger.info('Done.')

    elif (args.command == 'DEobsboot'):
        logger.info('Running DEobsboot command ...')
        m = importlib.import_module(name=f'pySpade.{args.command}')

        _ = m.DE_observe_bootstrap_cells(
            sub_df_file = args.transcriptome_df,
            sgrna_df = args.input_sgrna, 
            sgrnas_file = args.dict,
            num_processing = args.threads,
            norm = args.norm_method,
            output_dir = args.output_dir
            )        
        logger.info('Done.')

    elif (args.command == 'DErand'):
        logger.info('Running DErand command ...')
        m = importlib.import_module(name=f'pySpade.{args.command}')

        _ = m.DE_random_cells(
            sub_df_file = args.transcriptome_df,
            sgrna_df = args.input_sgrna,
            num_cell = args.num,
            iteration = args.iteration,
            num_processing = args.threads,
            norm = args.norm_method,
            RAND_M = args.randomization_method,
            output_dir = args.output_dir
            )        
        logger.info('Done.')

    elif (args.command == 'local'):
        logger.info('Running local command ...')
        m = importlib.import_module(name=f'pySpade.{args.command}')

        _ = m.local_analysis(
            FILE_DIR = args.file_dir,
            OBS_DIR = args.data_dir,
            DISTRI_DIR = args.distr,
            SGRNA_DICT = args.sgrna_dict,
            OUTPUT_DF = args.output_file
            )        
        logger.info('Done.')

    elif (args.command == 'global'):
        logger.info('Running global command ...')
        m = importlib.import_module(name=f'pySpade.{args.command}')

        _ = m.global_analysis(
            FILE_DIR = args.file_dir,
            OBS_DIR = args.data_dir,
            SGRNA_DICT = args.sgrna_dict,
            DISTRI_DIR = args.distr,
            OUTPUT_DF = args.output_file
            )        
        logger.info('Done.')

if __name__ == "__main__":
    main()


#global.py
import os
import re
import sys
import collections
import argparse
import tables
import itertools 
import scipy
import matplotlib
import csv
import glob

import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.sparse as sp_sparse
import scipy.io as sio

from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from multiprocessing import Pool
from collections import defaultdict
from scipy import sparse
from scipy.sparse import csr_matrix
from scipy.cluster.hierarchy import dendrogram, linkage

from pySpade.utils import get_logger, read_annot_df, read_sgrna_dict, load_data

logger = get_logger(logger_name=__name__)

def global_analysis(FILE_DIR,
                    OBS_DIR,
                    SGRNA_DICT,
                    DISTRI_DIR,
                    OUTPUT_DF):
    
    logger.info('Loading files.')
    #read the gene sequence file 
    gene_seq = np.load(FILE_DIR + 'Trans_genome_seq.npy', allow_pickle=True)
    if len(gene_seq) != len(set(gene_seq)):
        logger.critical('Duplication of mapping genes. Duplicates are removed in the analysis.')
    unique_elements, counts = np.unique(gene_seq, return_counts=True)
    duplicate_elements = unique_elements[counts > 1]
    #read the plotting annotation
    annot_df_dup = read_annot_df()
    #There are many non-coding genes duplication in the annot_df, only keep one.
    annot_df = annot_df_dup.drop_duplicates(subset='gene_names', keep='first')

    ##Load sgRNA dict: All regions 
    sgrna_dict  = read_sgrna_dict(SGRNA_DICT)

    #Save the randomized bin with list 
    
    Num = [int(i.split('/')[-1].split('.')[0].split('-')[-1]) for i in glob.glob(DISTRI_DIR+ '/Down_dist_mean-*')]
    if len([int(i.split('/')[-1].split('.')[0].split('-')[-1]) for i in glob.glob(DISTRI_DIR+ '/Down_dist_mean-*')]) != \
        len([int(i.split('/')[-1].split('.')[0].split('-')[-1]) for i in glob.glob(DISTRI_DIR+ '/Up_dist_mean-*')]):
        logger.critical('Background distribution files error!')
        sys.exit(0)

    logger.info(f'{len(Num)} sets of background distribution for global analysis.')

    #Calcualte pval
    df_column_list = [
        'idx', 'gene_names', 'chromosome', 'pos', 'strand', 
        'color_idx', 'chr_idx', 
        'region', 'num_cell', 'bin',
        'log(pval)-hypergeom', 'fc', 'log(padj)-Gaussian', 'fc_by_rand_dist_cpm', 'pval-empirical', 'cpm_perturb', 'cpm_bg']
    global_hits_df = pd.DataFrame(columns=df_column_list)

    #Generate bin-region dictionary 
    bin_dict = defaultdict(list)
    for region in list(sgrna_dict.keys()):
        fc_files = glob.glob(OBS_DIR + region + '*-foldchange')
        if len(fc_files) == 0:
            logger.critical('Cannot find foldchange file: ' + str(region))
            continue
        fc_file = fc_files[0]
        cell_num = int(fc_file.split('/')[-1].split('-')[-2])
        chosen_bin = Num[np.argmin([np.absolute(n-cell_num) for n in Num])]
        bin_dict[chosen_bin].append(region)
    logger.info('Finish generating bin-region dictionary')
    
    for b in Num:
        logger.info(f'Processing bin: {b}')
        logger.info(f'Num of regions in this bin: {len(bin_dict[b])}')

        #Load the background files
        down_mean = np.load(DISTRI_DIR + 'Down_dist_mean-%s.npy'%(str(b)))
        down_std = np.load(DISTRI_DIR + 'Down_dist_std-%s.npy'%(str(b))) 
        up_mean = np.load(DISTRI_DIR + 'Up_dist_mean-%s.npy'%(str(b)))
        up_std = np.load(DISTRI_DIR + 'Up_dist_std-%s.npy'%(str(b)))
        cpm_mean = np.load(DISTRI_DIR + 'Cpm_mean-%s.npy'%(str(b)))

        for region in bin_dict[b]:
            pval_list_up, pval_list_down, cpm, fc, cell_num = load_data(OBS_DIR, region)
            fc_cpm = (cpm + 0.01)/(cpm_mean + 0.01)

            #preprocess pval list 
            pval_list_up[np.argwhere(pval_list_up == 0)] = 1
            pval_list_up[np.isinf(pval_list_up)] = 0
            pval_list_up[np.isnan(pval_list_up)] = 0
            pval_list_down[np.argwhere(pval_list_down == 0)] = 1
            pval_list_down[np.isinf(pval_list_down)] = 0
            pval_list_down[np.isnan(pval_list_down)] = 0

            #split up-regulation and down-regulation
            pval_cutoff = 0
            fc_cutoff = 0
            up_i = np.where(np.array(fc_cpm) > (1 + fc_cutoff))[0]
            up_idx = np.array(list(set(up_i).intersection(set(np.where(pval_list_up < pval_cutoff)[0]))))

            down_i = np.where(np.array(fc_cpm) < (1 - fc_cutoff))[0]
            down_idx = np.array(list(set(down_i).intersection(set(np.where(pval_list_down < pval_cutoff)[0]))))
            #Calculate the overlap genes with annot_df, and only save the information on those genes.
            unique_elements, unique_indices = np.unique(gene_seq, return_index=True)

            up_keep_genes = list(set(annot_df['gene_names']).intersection(set(gene_seq[up_idx]))) #up-regulated genes found in both annotation file and transcriptome df
            up_keep_genes_idx = sorted(list(unique_indices[np.where(np.isin(unique_elements, up_keep_genes))[0]]))
            down_keep_genes = list(set(annot_df['gene_names']).intersection(set(gene_seq[down_idx])))
            down_keep_genes_idx = sorted(list(unique_indices[np.where(np.isin(unique_elements, down_keep_genes))[0]]))

            #Calculate p-value adj for Gaussian method 
            down_zscore_list = (pval_list_down[down_keep_genes_idx] - down_mean[down_keep_genes_idx]) / down_std[down_keep_genes_idx]
            down_padj_list = scipy.stats.norm.logsf(abs(down_zscore_list))
            down_hit_fc_list = fc_cpm[down_keep_genes_idx]
            down_cpm = cpm[down_keep_genes_idx]
            down_cpm_bg = cpm_mean[down_keep_genes_idx]

            up_zscore_list = (pval_list_up[up_keep_genes_idx] - up_mean[up_keep_genes_idx]) / up_std[up_keep_genes_idx]
            up_padj_list = scipy.stats.norm.logsf(abs(up_zscore_list))
            up_hit_fc_list = fc_cpm[up_keep_genes_idx]
            up_cpm = cpm[up_keep_genes_idx]
            up_cpm_bg = cpm_mean[up_keep_genes_idx]

            #Load background distribution and calculate emprical p-value
            rand_down_file = sio.loadmat(DISTRI_DIR + '%s-down_log-pval'%(str(b)))
            rand_down_matrix = []
            rand_down_matrix = sp_sparse.vstack(rand_down_file['matrix'])
            iter_num, gene_num = rand_down_matrix.shape
            emp_pval_down = np.sum(np.asarray(rand_down_matrix.tocsr()[:, down_keep_genes_idx].todense()) < pval_list_down[down_keep_genes_idx], axis=0) / iter_num

            rand_up_file = sio.loadmat(DISTRI_DIR + '%s-up_log-pval'%(str(b)))
            rand_up_matrix = []
            rand_up_matrix = sp_sparse.vstack(rand_up_file['matrix'])
            iter_num, gene_num = rand_up_matrix.shape
            emp_pval_up = np.sum(np.asarray(rand_up_matrix.tocsr()[:, up_keep_genes_idx].todense()) < pval_list_up[up_keep_genes_idx], axis=0) / iter_num

            #save to csv file: down-regulation gene 
            global_gene_series = annot_df[annot_df['gene_names'].isin(down_keep_genes)].set_index('idx').sort_index()
            global_gene_series['region'] = region
            global_gene_series['num_cell'] = cell_num
            global_gene_series['bin'] = b
            global_gene_series['log(pval)-hypergeom'] = pval_list_down[down_keep_genes_idx]
            global_gene_series['fc'] = fc[down_keep_genes_idx]
            global_gene_series['log(padj)-Gaussian'] = down_padj_list
            global_gene_series['fc_by_rand_dist_cpm'] = down_hit_fc_list
            global_gene_series['pval-empirical'] = emp_pval_down
            global_gene_series['cpm_perturb'] = down_cpm
            global_gene_series['cpm_bg'] = down_cpm_bg
            global_gene_series['idx'] = global_gene_series.index
            #global_hits_df = global_hits_df.append(global_gene_series)
            global_hits_df = pd.concat([global_hits_df, global_gene_series], ignore_index=True)

            #save to csv file: up-regulation gene 
            global_gene_series = annot_df[annot_df['gene_names'].isin(up_keep_genes)].set_index('idx').sort_index()
            global_gene_series['region'] = region
            global_gene_series['num_cell'] = cell_num
            global_gene_series['bin'] = b
            global_gene_series['log(pval)-hypergeom'] = pval_list_up[up_keep_genes_idx]
            global_gene_series['fc'] = fc[up_keep_genes_idx]
            global_gene_series['log(padj)-Gaussian'] = up_padj_list
            global_gene_series['fc_by_rand_dist_cpm'] = up_hit_fc_list
            global_gene_series['pval-empirical'] = emp_pval_up
            global_gene_series['cpm_perturb'] = up_cpm
            global_gene_series['cpm_bg'] = up_cpm_bg
            global_gene_series['idx'] = global_gene_series.index
            #global_hits_df = global_hits_df.append(global_gene_series)
            global_hits_df = pd.concat([global_hits_df, global_gene_series], ignore_index=True)
            
            logger.info(f'Finish region: {region}')
    
    global_hits_df = global_hits_df.reindex(columns=df_column_list)
    rem_dup_global_hits_df = global_hits_df[~global_hits_df['gene_names'].isin(duplicate_elements)]

    if OUTPUT_DF.endswith('.csv'):
        rem_dup_global_hits_df[rem_dup_global_hits_df['log(padj)-Gaussian'] < pval_cutoff].to_csv(OUTPUT_DF)
    else:
        rem_dup_global_hits_df[rem_dup_global_hits_df['log(padj)-Gaussian'] < pval_cutoff].to_csv(OUTPUT_DF + '.csv')

if __name__ == '__main__':
    pass



#DErand.py

import os
import sys
import re
import collections
import argparse
import tables
import itertools
import time
import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.sparse as sp_sparse
import glob

from multiprocessing import Pool
from collections import defaultdict
from scipy import sparse, io
from scipy.sparse import csr_matrix
from random import shuffle

from pySpade.utils import get_logger, read_annot_df, cpm_normalization, metacelll_normalization, perform_DE, hypergeo_test, get_num_processes

np.random.seed(0)
logger = get_logger(logger_name=__name__)

def DE_random_cells(sub_df_file,
                    sgrna_df,
                    num_cell,
                    RAND_M,
                    output_dir,
                    iteration=1000,
                    num_processing=1,
                    norm='cpm'):
    
    num_processing = get_num_processes()
    
    #check the normalization method
    if (norm != 'cpm') and (norm != 'metacell'):
        logger.critical("Incorrect normalization method. Has to be either 'cpm' or 'metacell'")
        sys.exit(0)

    #check the normalization method
    if (RAND_M != 'equal') and (RAND_M != 'sgrna'):
        logger.critical("Incorrect randomization method. Has to be either 'equal' or 'sgrna'")
        sys.exit(0)

    #read the 10X hdf5 file
    logger.info('Reading transcriptome file')
    if (sub_df_file.endswith('pkl')):
        sub_df = pd.read_pickle(sub_df_file)
    elif (sub_df_file.endswith('h5')):
        sub_df = pd.read_hdf(sub_df_file, 'df')
    
    #read the plotting annotation
    annot_df = read_annot_df()
    idx = np.arange(0, len(sub_df.index))
  
    #normalize the matrix.
    if (norm == 'cpm'):
        cpm_matrix = cpm_normalization(sub_df)
        
    elif (norm == 'metacell'):
        cpm_matrix = metacelll_normalization(sub_df)
     
    logger.info('Finished transcriptome normalization.')
    

    if (sgrna_df.endswith('pkl')):
        sgrna_df_adj_bool = pd.read_pickle(sgrna_df) > 0 
    elif (sgrna_df.endswith('h5')):
        sgrna_df_adj_bool = pd.read_hdf(sgrna_df, 'df') > 0

    [g,c] = sgrna_df_adj_bool.shape
    assert np.sum(sub_df.columns == sgrna_df_adj_bool.columns) == c
    if (RAND_M == 'sgrna'):
        total_choose_num = np.sum(np.sum(sgrna_df_adj_bool))     
        sgrna_weight_list = list(np.sum(sgrna_df_adj_bool, axis=0).values / total_choose_num)
    del sgrna_df_adj_bool
    logger.info('Finished processing sgRNA df.')

    combined_up_array = np.empty([iteration, len(sub_df.index)])
    combined_down_array = np.empty([iteration, len(sub_df.index)])
    combined_cpm_array = np.empty([iteration, len(sub_df.index)])
     
        
    cell_ID_dict = defaultdict(list)
    #idx index of cells containing the given sgRNA
    #random choose cells considering sgrna number per cell 
    logger.info('Start randomization Calculation.')
    for i in np.arange(iteration):
        counter = i
        if (counter % 50 == 0):
            logger.info(str(counter))

        if (RAND_M == 'sgrna'):
            sgrna_idx = np.random.choice(np.arange(c), size=num_cell, replace=False, p=sgrna_weight_list) #number (select from idx)
        elif (RAND_M == 'equal'):
            sgrna_idx = np.random.choice(np.arange(c), size=num_cell, replace=False)
                
        cell_ID_dict[i].append(list(sgrna_idx))
        #force the up-tail p-vals of all zero expressed genes to be zero. (actual p-val is 1)
        pval_list_down = np.zeros(len(sub_df.index))
        pval_list_up = np.zeros(len(sub_df.index))
        fc_list = np.ones(len(sub_df.index))
        cpm_list = np.zeros(len(sub_df.index))
            
        #perform the differential gene analysis by using Virtual FACS
        num_sgrna_cell, pval_list_up, pval_list_down, fc_list, cpm_list = perform_DE(
                                                                            sgrna_idx,
                                                                            cpm_matrix,
                                                                            idx,
                                                                            num_processing,
                                                                            pval_list_down,
                                                                            pval_list_up,
                                                                            fc_list,
                                                                            cpm_list
                                                                        )

        #force the up-tail p-vals of all zero expressed genes to be zero. (actual p-val is 1)
        combined_up_array[(counter)] = pval_list_up
        combined_down_array[(counter)] = pval_list_down
        combined_cpm_array[(counter)] = cpm_list
        
        #write the selected cell ID into file 
        output_file = open(output_dir + '/' + str(num_cell) + '-rand_cell_ID.txt', 'w')
        for i in cell_ID_dict.keys():
            select_cell_ID = cell_ID_dict[i]
            output_file.write(str(i) + '\t' + ','.join(str(j) for j in select_cell_ID) + '\n')
        output_file.close()

        up_pval_matrix   = csr_matrix(combined_up_array)
        down_pval_matrix = csr_matrix(combined_down_array)
        region_cpm_matrix = csr_matrix(combined_cpm_array)
   
        #save all the output
        io.savemat(
            '%s/%s-up_log-pval'%(output_dir, num_cell),
            {'matrix':up_pval_matrix}
        )

        io.savemat(
            '%s/%s-down_log-pval'%(output_dir, num_cell),
            {'matrix':down_pval_matrix}
        )

        io.savemat(
            '%s/%s-cpm'%(output_dir, num_cell),
            {'matrix':region_cpm_matrix}
        )

    #Process 1000 iteration
    logger.info('Start randomization summary.')
    up_distribution_mean = np.mean(np.array(up_pval_matrix.tocsr().todense()), axis=0)
    up_distribution_std = np.std(np.array(up_pval_matrix.tocsr().todense()), axis=0)
    down_distribution_mean = np.mean(np.array(down_pval_matrix.tocsr().todense()), axis=0)
    down_distribution_std = np.std(np.array(down_pval_matrix.tocsr().todense()), axis=0)
    cpm_mean = np.mean(np.array(region_cpm_matrix.tocsr().todense()), axis=0)
    
    up_01_perc = np.percentile(np.array(up_pval_matrix.tocsr().todense()), 0.1, axis=0)
    down_01_perc = np.percentile(np.array(down_pval_matrix.tocsr().todense()), 0.1, axis=0)

    np.save(output_dir + 'Up_dist_mean-%s'%(num_cell), up_distribution_mean)
    np.save(output_dir + 'Up_dist_std-%s'%(num_cell), up_distribution_std)
    np.save(output_dir + 'Down_dist_mean-%s'%(num_cell), down_distribution_mean)
    np.save(output_dir + 'Down_dist_std-%s'%(num_cell), down_distribution_std)
    np.save(output_dir + 'Cpm_mean-%s'%(num_cell), cpm_mean)

    np.save(output_dir + 'Up_01_perc-%s'%(num_cell), up_01_perc)
    np.save(output_dir + 'Down_01_perc-%s'%(num_cell), down_01_perc)

if __name__ == '__main__':
    pass



#utils.py
import logging
import numpy as np
import pandas as pd
import multiprocessing
from multiprocessing import Pool
import scipy.stats as stats
import scipy.sparse as sp_sparse
import collections
from collections import defaultdict
import itertools 
from importlib import resources
from scipy import sparse, io
from scipy.sparse import csr_matrix
import glob
import re
import tables
import matplotlib.pyplot as plt

def get_logger(logger_name, log_file=False):
    """Creates a custom logger."""

    FORMATTER = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    logger = logging.getLogger(logger_name)
    logger.setLevel(level=logging.DEBUG)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(logging.Formatter(FORMATTER))
    logger.addHandler(console_handler)

    if log_file:
        file_handler = logging.FileHandler(filename='pySpade.log')
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(FORMATTER))
        logger.addHandler(file_handler)

    return logger


def process_feature_df(mtx, barcodes, feature):
    gene_name = []
    for i in feature:
        gene_name.append(i[1])
    transcriptome_df = pd.DataFrame.sparse.from_spmatrix(data = mtx.tocsr(),
                                                         columns=barcodes, index=gene_name)
    return transcriptome_df

CountMatrix = collections.namedtuple('CountMatrix', ['feature_ref', 'barcodes', 'matrix'])
def get_matrix_from_h5(filename):
    with tables.open_file(filename, 'r') as f:
        mat_group = f.get_node(f.root, 'matrix')
        barcodes = f.get_node(mat_group, 'barcodes').read()
        data = getattr(mat_group, 'data').read()
        indices = getattr(mat_group, 'indices').read()
        indptr = getattr(mat_group, 'indptr').read()
        shape = getattr(mat_group, 'shape').read()
        matrix = sp_sparse.csc_matrix((data, indices, indptr), shape=shape)
         
        feature_ref = {}
        feature_group = f.get_node(mat_group, 'features')
        feature_ids = getattr(feature_group, 'id').read()
        feature_names = getattr(feature_group, 'name').read()
        feature_types = getattr(feature_group, 'feature_type').read()
        feature_ref['id'] = feature_ids
        feature_ref['name'] = feature_names
        feature_ref['feature_type'] = feature_types
        tag_keys = getattr(feature_group, '_all_tag_keys').read()
        for key in tag_keys:
            key = key.decode('utf-8')
            feature_ref[key] = getattr(feature_group, key).read()
         
        return CountMatrix(feature_ref, barcodes, matrix)

def turn_point(sgRNA_name, df):
    sgRNA_count  = df.T.filter(items=[sgRNA_name]).sum(axis=1).sort_values(ascending=False)
    sgRNA_cumsum = sgRNA_count.cumsum()

    #get the total cell number of this sgRNA
    cell_num = np.sum(sgRNA_count > 0)

    #calculate the turning point by using the max derivative
    turning_point = sgRNA_cumsum.loc[((sgRNA_cumsum.diff()) / sgRNA_count.sum() > (1/cell_num))].shape
    
    return(sgRNA_count.iloc[turning_point])

def filter_umi (df, copy=False):
    df = df.copy() if copy else df
    feature_cutoff = [[turn_point(i, df)] for i in list(df.index)]
    
    for i in range(0, len(feature_cutoff)):
        ZERO_HTO = df.iloc[i, :].loc[df.iloc[i, :] <= feature_cutoff[i][0]].index
        df.at[df.index[i], ZERO_HTO] = 0
    return df, feature_cutoff

def get_sgrna_per_cell(df, return_mean=True, return_median=True):
    CorrSgrnaPerCell = np.sum(df > 0, 0)
    sgrna_mean = np.mean(CorrSgrnaPerCell)
    sgrna_median = np.median(CorrSgrnaPerCell)
    
    if return_mean & return_median:
        return CorrSgrnaPerCell, sgrna_mean, sgrna_median
    elif return_mean:
        return CorrSgrnaPerCell, sgrna_mean
    elif return_median:
        return CorrSgrnaPerCell, sgrna_median
    else:
        return CorrSgrnaPerCell


def read_sgrna_dict(SGRNA_FILE):
    sgrna_dict  = {}
    with open(SGRNA_FILE) as f:
        for line in f:
            region_id, sgrna_string = line.strip().split("\t")
            sgrnas = sgrna_string.split(";")
            sgrna_dict.update({region_id : sgrnas})
    return(sgrna_dict)

def read_annot_df() -> pd.DataFrame:
    with resources.path('pySpade', 'plot_annotation.txt') as df:
        annot_df = pd.read_csv(df,
                   header=None,
                   sep='\t',
                   names=['idx', 'gene_names', 'chromosome', 'pos', 'strand', 'color_idx', 'chr_idx'])
        return annot_df

#be sure that when calculating find_sgrna_cells, the column sequences are the same between transcriptome df and sgrna df
#modified, specific for single sgRNA
def find_single_sgrna_cells(sgRNA_df, sgrna):
    cell_bc = sgRNA_df.loc[:, (sgRNA_df.loc[sgrna] > 0)].T.index
    cell_index = []
    for i in cell_bc:
        current_idx = np.argwhere(sgRNA_df.columns == i)
        if  current_idx.size > 0:
            cell_index.append(current_idx.item())
    return [x for x in set(cell_index)]

def find_all_sgrna_cells(sgRNA_df, sgRNA_dict):
    cell_bc = sgRNA_df.loc[:, (sgRNA_df.loc[sgRNA_dict].sum() > 0)].T.index
    cell_index = []
    for i in cell_bc:
        current_idx = np.argwhere(sgRNA_df.columns == i)
        if  current_idx.size > 0:
            cell_index.append(current_idx.item())
    return [x for x in set(cell_index)]

def cpm_normalization(trans_df):
    [g,c] = trans_df.shape
    cpm_matrix = np.zeros((g,c))
    cpm_matrix = np.array(trans_df / trans_df.sum(axis = 0) * 1e6)
    return cpm_matrix

def metacelll_normalization(trans_df):
    [g,c] = trans_df.shape
    cpm_matrix = np.zeros((g,c))
    uniq_id = set()
    for x in trans_df.columns:
        uniq_id.add(x[-1])
    for lib in sorted(uniq_id):
        index = [i for i,e in enumerate(trans_df.columns) if e.split('-')[1] == lib]
        one_cell_cpm = np.array((trans_df.iloc[:,index].sum(axis = 1) + 1)\
                        / np.sum(trans_df.iloc[:,index].values) * 1e6).flatten()
        for i in index:
            cell_cpm = trans_df.iloc[:,i] / np.sum(trans_df.iloc[:,i].values) * 1e6
            cpm_matrix[:,i] = cell_cpm / one_cell_cpm  

def set_axis_style(ax, labels):
    ax.xaxis.set_tick_params(direction='out')
    ax.xaxis.set_ticks_position('bottom')
    ax.set_xticks(np.arange(1, len(labels) + 1))
    ax.set_xticklabels(labels, fontsize=14, color='#000000')
    ax.set_xlim(0.25, len(labels) + 0.75)

def sc_exp_box_plot(sgrna_cells_expression, other_cells_expression, target_gene, region, output_file):
    fig, ax = plt.subplots(figsize= (8,6))
    ax.set_title(target_gene + ' expression in ' + region, fontsize=20)
    flierprops = dict(marker='o', markerfacecolor='#000000', markersize=6, alpha=0.5,
                    linestyle='none')
    ax.boxplot([sgrna_cells_expression, other_cells_expression], whis=[10, 90],
            showfliers=True, flierprops=flierprops)
    labels = ['sgRNA cells', 'other cells']
    set_axis_style(ax, labels)
    plt.savefig(output_file)

def get_num_processes():
    # Adjust this logic as needed based on your requirements
    return min(multiprocessing.cpu_count(), 100)  # Set an upper limit, e.g., 4 processes

def hypergeo_test(non_zero_array, sgrna_idx, i):
    #find indecies of cells in which expression of given gene is
    #equal or less than the median of this gene in the whole population
    median_cell_idx  = np.argwhere(non_zero_array <= np.median(non_zero_array))

    #find the same cells subset in the cells with a given sgRNA
    overlap_cell_idx = np.intersect1d(median_cell_idx, sgrna_idx)
    
    #calculate the median fold change
    other_idx = np.setxor1d(sgrna_idx, range(len(non_zero_array)))
    
    fc = (np.mean(non_zero_array[sgrna_idx]) + 0.01) / (np.mean(non_zero_array[other_idx]) + 0.01)
    cpm = np.mean(non_zero_array[sgrna_idx])
    
    #perform hypergeometric test, get the upper tail
    k = len(overlap_cell_idx)
    M = len(non_zero_array)
    n = len(median_cell_idx)
    N = len(sgrna_idx)
    try:
        pval_up = stats.hypergeom.logcdf(k, M, n, N).item()
    except:
        pval_up = float('nan')
        
    try:
        pval_down = stats.hypergeom.logsf(k, M, n, N).item()
    except:
        pval_down = float('nan')
    
    return pval_down, pval_up, fc, cpm

def perform_DE(sgrna_idx, input_array, idx, num_processes, pval_list_down, pval_list_up, fc_list, cpm_list):
    nonzero_pval_list_up = []
    nonzero_pval_list_down = []
    nonzero_fc_list = []
    nonzero_cpm_list = []
    
    with Pool(processes=num_processes) as p:
        for pval_down, pval_up, fc, cpm in p.starmap(hypergeo_test, zip(
                input_array,
                itertools.repeat(sgrna_idx),
                idx)
        ):
            nonzero_pval_list_down.append(pval_down)
            nonzero_pval_list_up.append(pval_up)
            nonzero_fc_list.append(fc)
            nonzero_cpm_list.append(cpm)
    for i in idx:
        pval_list_up[i] = nonzero_pval_list_up.pop(0)
        pval_list_down[i] = nonzero_pval_list_down.pop(0)
        fc_list[i] = nonzero_fc_list.pop(0)
        cpm_list[i] = nonzero_cpm_list.pop(0)
    return len(sgrna_idx), pval_list_up, pval_list_down, fc_list, cpm_list

def load_data(data_dir, region):
    up_pval_file   = data_dir + region + '-up_log-pval'
    down_pval_file = data_dir + region + '-down_log-pval'
    cpm_file = data_dir + region + '-cpm'
    fc_files = glob.glob(data_dir + region + '*-foldchange')
    if len(fc_files) == 1:
        fc_file = fc_files[0]
    else:
        numbers = np.array([int(float(i.split('/')[-1].split('-')[-2])) for i in fc_files])
        chosen_num = np.max(numbers)
        fc_file = data_dir + region + '-' + str(chosen_num) + '-foldchange'
    
    pval_list_up = io.loadmat(up_pval_file)['matrix'][0]
    pval_list_down = io.loadmat(down_pval_file)['matrix'][0]
    cpm = io.loadmat(cpm_file)['matrix'][0]
    fc = io.loadmat(fc_file)['matrix'][0]
    
    num_sgrna_cell = int(fc_file.split('/')[-1].split('-')[-2])
    return pval_list_up, pval_list_down, cpm, fc, num_sgrna_cell

def get_neighbor_genes(region, query_range, annot_df):
    
    length_list = [
        0, 248956422,491149951,689445510,879660065,1061198324,
        1232004303,1391350276,1536488912,1674883629,1808681051,
        1943767673,2077042982,2191407310,2298451028,2400442217,
        2490780562,2574038003,2654411288,2713028904,2777473071,
        2824183054,2875001522,3031029399
    ]
    
    chr_order = [
        'chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8',
        'chr9', 'chr10', 'chr11', 'chr12', 'chr13', 'chr14', 'chr15',
        'chr16', 'chr17', 'chr18', 'chr19', 'chr20', 'chr21', 'chr22', 'chrX'
    ]

    enh_chrom, left, right = re.split('[:|-]', region)
    position = int(left) + length_list[chr_order.index(enh_chrom)]
    gene = annot_df.loc[(annot_df.pos < position + query_range) \
                            & (annot_df.pos > position - query_range)\
                            & (annot_df.chromosome == enh_chrom)].gene_names.values
    return(gene)

def get_distance(region, gene_pos):
    length_list = [
        0, 248956422,491149951,689445510,879660065,1061198324,               
        1232004303,1391350276,1536488912,1674883629,1808681051,
        1943767673,2077042982,2191407310,2298451028,2400442217,
        2490780562,2574038003,2654411288,2713028904,2777473071,
        2824183054,2875001522,3031029399
    ]
    
    chr_order = [
        'chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 
        'chr9', 'chr10', 'chr11', 'chr12', 'chr13', 'chr14', 'chr15',
        'chr16', 'chr17', 'chr18', 'chr19', 'chr20', 'chr21', 'chr22', 'chrX'
    ]
    
    enh_chrom, left, right = re.split('[:|-]', region)
    dist =  int(gene_pos) - int(left) - length_list[chr_order.index(enh_chrom)]
    
    return np.absolute(dist)


#DEobs.py

import os
import sys
import re
import collections
import argparse
import tables
import itertools
import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.sparse as sp_sparse
import multiprocessing 
from multiprocessing import Pool
from collections import defaultdict
from scipy import sparse, io
from scipy.sparse import csr_matrix
from random import shuffle

from pySpade.utils import get_logger, read_annot_df, cpm_normalization, metacelll_normalization, read_sgrna_dict, find_all_sgrna_cells, perform_DE, hypergeo_test, get_num_processes

np.random.seed(0)
logger = get_logger(logger_name=__name__)

def DE_observe_cells(sub_df_file,
                    sgrna_df,
                    sgrnas_file,
                    output_dir,
                    num_processing=1,
                    norm='cpm',
                    ):
    
    num_processing = get_num_processes()

    #check the normalization method   
    if (norm != 'cpm') and (norm != 'metacell'):
        logger.critical("Incorrect normalization method. Has to be either 'cpm' or 'metacell'")
        sys.exit(0)

    #read the 10X hdf5 file
    logger.info('Reading transcriptome file')

    if (sub_df_file.endswith('pkl')):
        sub_df = pd.read_pickle(sub_df_file)
    elif (sub_df_file.endswith('h5')):
        sub_df = pd.read_hdf(sub_df_file, 'df')
       
    idx = np.arange(0, len(sub_df.index))

    #normalize the matrix.       
    if (norm == 'cpm'):
        cpm_matrix = cpm_normalization(sub_df)
        
    elif (norm == 'metacell'):
        cpm_matrix = metacelll_normalization(sub_df)
     
    logger.info('Finished transcriptome normalization.')
    

    #load the sgRNA file
    logger.info('Loading sgRNA df.')
    if (sgrna_df.endswith('pkl')):
        sgrna_df_adj_bool = pd.read_pickle(sgrna_df) > 0 
    elif (sgrna_df.endswith('h5')):
        sgrna_df_adj_bool = pd.read_hdf(sgrna_df, 'df') > 0

    [g,c] = sub_df.shape
    assert np.sum(sub_df.columns == sgrna_df_adj_bool.columns) == c #make sure the sequences are the same compare with two df
    
    del sub_df
    #perform hypergeometric test for every single gene in the dataframe
    sgrna_dict  = read_sgrna_dict(sgrnas_file)

    logger.info('Start DE analysis.')
    for k in sgrna_dict:
        logger.info(f'Start analysis of region: {k}')

        #Find if any sgRNA is missing from the sgRNA df
        missing_sgrna_list = list(set(sgrna_dict[k]) - set(sgrna_df_adj_bool.index))
        if len(missing_sgrna_list) == len(sgrna_dict[k]):
            logger.info('Missing all the sgRNA in this region.')
            continue
        if len(missing_sgrna_list) > 0:
            left_sgrna = len(sgrna_dict[k]) - len(missing_sgrna_list)
            logger.info('Missing ' + str(len(missing_sgrna_list)) + ' sgrna, ' + str(left_sgrna) + ' left for analysis.')
            for i in missing_sgrna_list:
                logger.info('Missing sgrna: ' + str(i))

        #remove the missing sgrna and continue DE analysis 
        if len(missing_sgrna_list) > 0:
            remain_sgrna_list = list(set(sgrna_dict[k]) - (set(missing_sgrna_list)))
            sgrna_dict.update({k:remain_sgrna_list})

        #idx index of cells containing the given sgRNA
        sgrna_idx = find_all_sgrna_cells(sgrna_df_adj_bool, sgrna_dict[k])
        if len(sgrna_idx) == 0:
            logger.info('No cell in this region.')
            continue
        #force the up-tail p-vals of all zero expressed genes to be zero. (actual p-val is 1)
        pval_list_down = np.zeros(g)
        pval_list_up = np.zeros(g)

        fc_list = np.ones(g)
        cpm_list = np.zeros(g)
       
        #perform the differential gene analysis
        num_sgrna_cell, pval_list_up, pval_list_down, fc_list, cpm_list = perform_DE(
            sgrna_idx,
            cpm_matrix,
            idx,
            num_processing,
            pval_list_down,
            pval_list_up,
            fc_list,
            cpm_list)
        
        #save all the output
        io.savemat(
            '%s/%s-up_log-pval'%(output_dir, k[0:]),
            {'matrix':pval_list_up})

        io.savemat(
            '%s/%s-down_log-pval'%(output_dir, k[0:]),
            {'matrix':pval_list_down})

        io.savemat('%s/%s-%s-foldchange'%(output_dir, k[0:], num_sgrna_cell),
                   {'matrix':fc_list})
        
        io.savemat(
            '%s/%s-cpm'%(output_dir, k[0:]),
            {'matrix':cpm_list})
   
if __name__ == '__main__':
    pass



#process.py

from asyncio.log import logger
import os
import re
import sys
import collections
import argparse
import tables
import itertools
import matplotlib
import numba

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.sparse as sp_sparse

from multiprocessing import Pool
from collections import defaultdict
from scipy import sparse, io
from scipy.sparse import csr_matrix
from pySpade.utils import get_logger, get_matrix_from_h5, process_feature_df, turn_point, filter_umi, get_sgrna_per_cell

logger = get_logger(logger_name=__name__)

def process_singlet(WORK_DIR,
                    SGRNA,
                    CELL_MULTIPLEX,
                    OUTPUT_DIR,
                    COMP='False'):
 
    logger.info('Reading data.')

    if (SGRNA.split('.')[-2] != 'pkl') and (SGRNA.split('.')[-1] != 'pkl') and (SGRNA.split('.')[-2] != 'csv') and (SGRNA.split('.')[-1] != 'csv'):
        logger.critical('sgRNA matrix format incorrect')
        sys.exit(0)

    if (COMP != 'True') and (COMP != 'False'):
        logger.critical("Incorrect compression method.")
        sys.exit(0)
    
    if (CELL_MULTIPLEX) != None:
        logger.info('Remove doublets in this dataset.')
        #calculate HTO and filter
        ab_list = []
        with open (CELL_MULTIPLEX) as f:
            for line in f:
                ab = line.strip()
                ab_list.append(ab)
        logger.info('Number of antibodies: ' + str(len(ab_list)))
        
        Count_list = []
        CountMatrix = collections.namedtuple('CountMatrix', ['feature_ref', 'barcodes', 'matrix'])
        filtered_matrix_h5 = WORK_DIR + 'filtered_feature_bc_matrix.h5'
        filtered_feature_bc_matrix = get_matrix_from_h5(filtered_matrix_h5)
        ab_idx = np.where(filtered_feature_bc_matrix.feature_ref['feature_type'] == b'Antibody Capture')[0]
        if len(ab_idx) != len(ab_list):
            logger.critical('Input Antibody list is not consistent with mapping info. Using mapping info for further analysis.')
            logger.critical(str(len(ab_idx)) + ' antibodies in the transcriptome dataframe.')
            logger.critical(str(len(ab_list) + ' antibodies provided in the list.'))
        
        #slice antibody matrix to filter UMI
        df = pd.DataFrame(data=filtered_feature_bc_matrix.matrix[ab_idx, :].todense(), \
                        columns=filtered_feature_bc_matrix.barcodes.astype(str), \
                        index=filtered_feature_bc_matrix.feature_ref['name'][ab_idx].astype(str))
        HTO_df_adj, _ = filter_umi(df, copy=True)
        del df
        Count_list.append(np.sum(HTO_df_adj > 0, axis=0).values)

        HTO_df_adj_bool = HTO_df_adj > 0
        del HTO_df_adj
        singlet_HTO_df = HTO_df_adj_bool.T[(HTO_df_adj_bool.sum(axis=0).values == 1)]

        singlet_rate = collections.Counter(Count_list[0])[1] / len(filtered_feature_bc_matrix.barcodes)
        logger.info('Finished processing experimental doublets.')

        #Load sgrna df 
        if (SGRNA.endswith('pkl')) | (SGRNA.endswith('pkl.gz')):
            sgRNA_df = pd.read_pickle(SGRNA)
        elif (SGRNA.endswith('csv')) | (SGRNA.endswith('csv.gz')):
            sgRNA_df = pd.read_csv(SGRNA, index_col=0)


        All_singlet_ID = HTO_df_adj_bool.columns[(HTO_df_adj_bool > 0).sum(axis=0).values == 1].values
        new_columns = set(sgRNA_df.columns.values).intersection(set(All_singlet_ID))

        new_sgRNA_df = sgRNA_df[new_columns]
        del sgRNA_df
        logger.info('Generated sgRNA matrix without experimental doublets.')

        #filter sgRNA outliers 
        CorrSgrnaPerCell, sgrna_mean , sgrna_median = get_sgrna_per_cell(new_sgRNA_df, return_mean=True, return_median=True)
        Q1 = CorrSgrnaPerCell.quantile(0.25)
        Q3 = CorrSgrnaPerCell.quantile(0.75)
        IQR = Q3 - Q1
        
        #generate new singlet df
        no_outlier_sgrna_df = new_sgRNA_df.loc[:,((CorrSgrnaPerCell < (Q1 - 1.5 * IQR)) |(CorrSgrnaPerCell > (Q3 + 1.5 * IQR))).values == False]
        del new_sgRNA_df
        CorrSgrnaPerCell_out, sgrna_mean_out , sgrna_median_out = get_sgrna_per_cell(no_outlier_sgrna_df, return_mean=True, return_median=True)
        
        logger.info('Generated transcriptome and sgRNA matrix without sgRNA outlier cells.')
        logger.info('singlet rate: ' + str(singlet_rate))
        logger.info('Cells before filter experimental doublets: ' + str(len(filtered_feature_bc_matrix.barcodes)))
        logger.info('Cells after filter experimental doublets: ' + str(len(new_columns)))
        logger.info('sgRNA mean before filter sgRNA outlier: ' + str(sgrna_mean))
        logger.info('sgRNA median before filter sgRNA outlier: ' + str(sgrna_median))
        logger.info('sgRNA mean after filter sgRNA outlier: ' + str(sgrna_mean_out))
        logger.info('sgRNA median after filter sgRNA outlier: ' + str(sgrna_median_out))
        logger.info('Cells after filter experimental doublets and sgRNA outliers: ' + str(len(no_outlier_sgrna_df.columns)))

        #write output matrix
        sub_df = pd.DataFrame(data=filtered_feature_bc_matrix.matrix.todense(), columns=filtered_feature_bc_matrix.barcodes.astype(str), index=filtered_feature_bc_matrix.feature_ref['name'].astype(str))
        np.save(OUTPUT_DIR + 'Trans_genome_seq', sub_df.index)
        np.save(OUTPUT_DIR + 'Perc_cell_expr', (np.sum(sub_df[no_outlier_sgrna_df.columns] > 0, axis=1) / len(no_outlier_sgrna_df.columns) *100).values)
        
        if (COMP=='True'):
            no_outlier_sgrna_df.to_hdf(OUTPUT_DIR + '/Singlet_sgRNA_df.h5', key='df', mode='w', complevel=9, complib='zlib')    
            sub_df[no_outlier_sgrna_df.columns].to_hdf(OUTPUT_DIR + '/Singlet_sub_df.h5', key='df', mode='w', complevel=9, complib='zlib')
        elif (COMP=='False'):
            no_outlier_sgrna_df.to_hdf(OUTPUT_DIR + '/Singlet_sgRNA_df.h5', key='df', mode='w')    
            sub_df[no_outlier_sgrna_df.columns].to_hdf(OUTPUT_DIR + '/Singlet_sub_df.h5', key='df', mode='w')
    
    #Without CellPlex, save every cell
    else:
        logger.info('No CellPlex file. Will not remove any doublets. Keep all the cells.')

        #Load sgrna df 
        if (SGRNA.endswith('pkl')) | (SGRNA.endswith('pkl.gz')):
            sgRNA_df = pd.read_pickle(SGRNA)
        elif (SGRNA.endswith('csv')) | (SGRNA.endswith('csv.gz')):
            sgRNA_df = pd.read_csv(SGRNA, index_col=0)

        #process transcriptome file 
        CountMatrix = collections.namedtuple('CountMatrix', ['feature_ref', 'barcodes', 'matrix'])
        filtered_matrix_h5 = WORK_DIR + 'filtered_feature_bc_matrix.h5'
        filtered_feature_bc_matrix = get_matrix_from_h5(filtered_matrix_h5)
        sub_df = pd.DataFrame(data=filtered_feature_bc_matrix.matrix.todense(), columns=filtered_feature_bc_matrix.barcodes.astype(str), index=filtered_feature_bc_matrix.feature_ref['name'].astype(str))
        
        new_columns = set(sgRNA_df.columns).intersection(set(sub_df.columns))
        CorrSgrnaPerCell_out, sgrna_mean , sgrna_median = get_sgrna_per_cell(sgRNA_df[new_columns], return_mean=True, return_median=True)
        logger.info('Cell number: ' + str(len(new_columns)))
        logger.info('sgRNA mean: ' + str(sgrna_mean))
        logger.info('sgRNA median: ' + str(sgrna_median))


        #write output matrix
        np.save(OUTPUT_DIR + 'Trans_genome_seq', sub_df.index)
        np.save(OUTPUT_DIR + 'Perc_cell_expr', (np.sum(sub_df[new_columns] > 0, axis=1) / len(new_columns) *100).values)
        
        if (COMP=='True'):
            sgRNA_df[new_columns].to_hdf(OUTPUT_DIR + '/Singlet_sgRNA_df.h5', key='df', mode='w', complevel=9, complib='zlib')    
            sub_df[new_columns].to_hdf(OUTPUT_DIR + '/Singlet_sub_df.h5', key='df', mode='w', complevel=9, complib='zlib')
        elif (COMP=='False'):
            sgRNA_df[new_columns].to_hdf(OUTPUT_DIR + '/Singlet_sgRNA_df.h5', key='df', mode='w')    
            sub_df[new_columns].to_hdf(OUTPUT_DIR + '/Singlet_sub_df.h5', key='df', mode='w')    

if __name__ == '__main__':
    pass



from setuptools import setup, find_packages
from pySpade import __version__

with open(file='README.md', mode='r') as fh:
    long_description = fh.read()

    setup(
    name='pySpade',
    version=__version__,
    author='Yihan Wang',
    author_email='Yihan.Wang@UTSouthwestern.edu',
    description='Single cell Perturbations - Analysis of Differential gene Expression',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Hon-lab/pySpade',
    packages=find_packages(exclude=('docs', 'tests')),
    package_dir={'pySpade': 'pySpade'},
    package_data={'pySpade': ['*.txt']},
    include_package_data=True,
    classifiers=[
                'Programming Language :: Python :: 3',
                'License :: OSI Approved :: MIT License',
                'Operating System :: OS Independent'
    ],
    python_requires='>=3.7',
    entry_points={
        'console_scripts': ['pySpade=pySpade.__main__:main'],
    }
)



#local.py

import os
import re
import sys
import collections
import argparse
from threading import local
import tables
import itertools 
import scipy
import csv
import glob

import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.sparse as sp_sparse
import scipy.io as sio

from collections import defaultdict
from scipy import sparse
from scipy.sparse import csr_matrix

from pySpade.utils import get_logger, read_annot_df, get_neighbor_genes, get_distance, read_sgrna_dict

logger = get_logger(logger_name=__name__)

def local_analysis(FILE_DIR,
                    OBS_DIR,
                    DISTRI_DIR,
                    SGRNA_DICT,
                    OUTPUT_DF):
    
    logger.info('Loading files.')

    #read the gene sequence file 
    gene_seq = np.load(FILE_DIR + 'Trans_genome_seq.npy', allow_pickle=True)
    if len(gene_seq) != len(set(gene_seq)):
        logger.critical('Duplication of mapping genes. Duplicates are removed in the analysis.')
    unique_elements, counts = np.unique(gene_seq, return_counts=True)
    duplicate_elements = unique_elements[counts > 1]
 
        
    #read the plotting annotation
    annot_df_dup = read_annot_df()
    #There are many non-coding genes duplication in the annot_df, only keep one.
    annot_df = annot_df_dup.drop_duplicates(subset='gene_names', keep='first')
    
    #Load sgRNA dict: All regions 
    sgrna_dict  = read_sgrna_dict(SGRNA_DICT)

    #read all the perturbation files
    pval_files = glob.glob(OBS_DIR + '*-down_log-pval')

    df_column_list = [
        'gene_names', 'chromosome', 'pos', 'strand', 
        'color_idx', 'chr_idx', 'region', 'distance', 'num_cell', 'bin', 
        'log(pval)-hypergeom', 'fc', 'log(padj)-Gaussian', 'fc_by_rand_dist_cpm', 'pval-empirical', 'cpm_perturb', 'cpm_bg']
    local_gene_df = pd.DataFrame(columns=df_column_list)

    #Read the background distribution file
    Num = [int(i.split('/')[-1].split('.')[0].split('-')[-1]) for i in glob.glob(DISTRI_DIR+ '/Down_dist_mean-*')]
    if len([int(i.split('/')[-1].split('.')[0].split('-')[-1]) for i in glob.glob(DISTRI_DIR+ '/Down_dist_mean-*')]) != \
        len([int(i.split('/')[-1].split('.')[0].split('-')[-1]) for i in glob.glob(DISTRI_DIR+ '/Up_dist_mean-*')]):
        logger.critical('Background distribution files error!')
        sys.exit(0)

    #start calculating local hits for each perturbation region
    logger.info('Start analysis of ' + str(len(pval_files)) + ' regions.')
    for region in list(sgrna_dict.keys()):
        logger.info(f'  Processing region: {region}')
        if region.startswith('chr') == False:   
            logger.info('No chromosome coordination in this region, cannot compute local analysis.')
            continue

        cpm_file = OBS_DIR + region + '-cpm'
        cpm = sio.loadmat(cpm_file)['matrix'][0]        
        fc_files = glob.glob(OBS_DIR + region + '*-foldchange')
        if len(fc_files) == 1:
            fc_file = fc_files[0]
            cell_num = int(fc_file.split('/')[-1].split('-')[-2])
        else:    
            numbers = np.array([int(float(i.split('/')[-1].split('-')[2])) for i in fc_files])
            chosen_num = np.max(numbers)
            cell_num = int(chosen_num)
            fc_file = OBS_DIR + region + '-' + str(chosen_num) + '-foldchange'
        
        fc = sio.loadmat(fc_file)['matrix'][0]  #compare to all the other cells as background

        #get the gene idx within local analysis window and filter with fold change 
        local_gene = get_neighbor_genes(region, 2e6, annot_df)
        fc_cutoff = 0.01
        pval_cutoff = 0

        #read the pval matrix and process
        pval = sio.loadmat(OBS_DIR + region + '-down_log-pval')['matrix'][0] #raw hypergeom p value
        pval[np.argwhere(pval == 0)] = 1
        pval[np.isinf(pval)] = 0
        pval[np.isnan(pval)] = 0

        #Calculate the overlap genes with annot_df, and only save the information on those genes.
        down_i = np.where(np.array(fc) < (1 - fc_cutoff))[0]
        down_idx = np.array(list(set(down_i).intersection(set(np.where(pval < pval_cutoff)[0]))))
        ##Remove the duplicate value of gene index, return a list with unique gene names. 
        ##Unique_elements is sorted by the gene names, unique_indices return the original index from gene_seq.
        unique_elements, unique_indices = np.unique(gene_seq, return_index=True)
        down_keep_genes = list((set(annot_df['gene_names']).intersection(set(gene_seq[down_idx]))).intersection(set(local_gene)))
        down_keep_genes_idx = sorted(list(unique_indices[np.where(np.isin(unique_elements, down_keep_genes))[0]]))
        #Make this true: gene_seq[down_keep_genes_idx] == down_keep_genes

        if len(down_keep_genes_idx) ==0:
            logger.info(f'  No down-regulation genes within local analysis windown. ')
            continue
        
        #Calculate the adjusted pval with closet cell number distribution
        chosen_dist = Num[np.argmin([np.absolute(n - cell_num) for n in Num])]
        
        #Load the file
        down_mean = np.load(DISTRI_DIR + 'Down_dist_mean-%s.npy'%(str(chosen_dist)))
        down_std = np.load(DISTRI_DIR + 'Down_dist_std-%s.npy'%(str(chosen_dist)))
        cpm_mean = np.load(DISTRI_DIR + 'Cpm_mean-%s.npy'%(str(chosen_dist)))
        fc_rand = cpm/cpm_mean
        
        #Calculate adjusted p value of Gaussian padj
        down_zscore_list = (pval[down_keep_genes_idx] - down_mean[down_keep_genes_idx]) / down_std[down_keep_genes_idx]
        down_padj_list = scipy.stats.norm.logsf(abs(down_zscore_list))
        hits_fc = fc[down_keep_genes_idx]
        hits_fc_rand = fc_rand[down_keep_genes_idx]

        #Load pvalue matrix and calculate empirical p-value
        rand_down_file = sio.loadmat(DISTRI_DIR + '%s-down_log-pval'%(str(chosen_dist)))
        rand_down_matrix = []
        rand_down_matrix = sp_sparse.vstack(rand_down_file['matrix'])
        iter_num, gene_num = rand_down_matrix.shape
        emp_pval = np.sum(np.asarray(rand_down_matrix.tocsr()[:, down_keep_genes_idx].todense()) < pval[down_keep_genes_idx], axis=0) / iter_num

        #Save to csv file 
        local_gene_series = annot_df[annot_df['gene_names'].isin(down_keep_genes)].set_index('idx').sort_index()
        local_gene_series['region'] = region
        dist_list = []
        for i in local_gene_series['pos']:
            dist = get_distance(region, i)
            dist_list.append(dist)
        local_gene_series['distance'] = dist_list
        local_gene_series['num_cell'] = cell_num
        local_gene_series['bin'] = chosen_dist
        local_gene_series['log(pval)-hypergeom'] = pval[down_keep_genes_idx] 
        local_gene_series['fc'] = hits_fc
        local_gene_series['log(padj)-Gaussian'] = down_padj_list
        local_gene_series['fc_by_rand_dist_cpm'] = hits_fc_rand
        local_gene_series['pval-empirical'] = emp_pval
        local_gene_series['cpm_perturb'] = cpm[down_keep_genes_idx]
        local_gene_series['cpm_bg'] = cpm_mean[down_keep_genes_idx]
        #local_gene_df = local_gene_df.append(local_gene_series)
        local_gene_df = pd.concat([local_gene_df, local_gene_series], ignore_index=True)
        local_gene_df = local_gene_df.reindex(columns=df_column_list)

    if OUTPUT_DF.endswith('.csv'):
        local_gene_df[~local_gene_df['gene_names'].isin(duplicate_elements)].to_csv(OUTPUT_DF)
    else:
        local_gene_df[~local_gene_df['gene_names'].isin(duplicate_elements)].to_csv(OUTPUT_DF + '.csv')

if __name__ == '__main__':
    pass



#parsers.py

import sys
import argparse
from pySpade import __version__

def parse_args(args=sys.argv[1:]):

    # create top-level parser
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=('pySpade \n' + "Version: " +
                     __version__))

    # create sub-level parser
    subparsers = parser.add_subparsers(title='functions',
                                       dest='command',
                                       metavar='')

    add_process_subparser(subparsers)
    add_explevel_subparser(subparsers)
    add_fc_subparser(subparsers)
    add_DEobs_subparser(subparsers)
    add_DEobsboot_subparser(subparsers)
    add_DErand_subparser(subparsers)
    add_local_subparser(subparsers)
    add_global_subparser(subparsers)

    if len(sys.argv) > 1:
        if (sys.argv[1] == '--version' or sys.argv[1] == '-v'):
            print(" version: %s" % __version__)
            exit()
    else:
        args = parser.parse_args(['-h'])
        exit()

    return parser.parse_args(args)

# process
def add_process_subparser(subparsers):
    parser = subparsers.add_parser(
        'process',
        help='process mapping output and reformat for downstream analysis.',
        description=(
            'Process transcriptome output and sgrna output to remove experimental doublets. '
            'Transcriptome matrix is from Cellranger output (outs folder),'
            'sgrna matrix is from fba output, other accepted format include pkl and csv file.'
            'sgrna matrix column: cell barcodes consistent with transcriptome matrix, rows: sgrna sequence.'
            'The final output format is h5 file.'))

    parser.add_argument('-f', 
                        '--feature_bc', 
                        dest='feature_bc', 
                        required=True,
                        type=str,
                        help='Specify the output folder from cellranger pipeline (outs folder)')

    parser.add_argument('-s', 
                        '--sgrna',
                        dest='input_sgrna',
                        required=True,
                        type=str,
                        help='Specify the sgrna matrix file.'
                        'Please make sure the barcodes are consistent with transcriptome if thre are multiple libraries. File format: pkl or csv')

    parser.add_argument('-m', 
                        '--cell_multiplex',
                        dest='cell_multiplex',
                        required=False,
                        default=False,
                        help='specify the used antibody name in txt file separate with new line. Make sure that the antibody is mapped together with transcriptome. Cellranger is recommended.')

    parser.add_argument('-o', 
                        '--output_directory',
                        dest='output_directory', 
                        required=True,
                        help='Specify output directory. The output file format will be h5')

    parser.add_argument('-c', 
                        '--comp', 
                        dest = 'comp', 
                        required=False,
                        type=str,
                        default='False',
                        help='Final h5 file compress or not. "True" or "False", default is False.')

# explevel
def add_explevel_subparser(subparsers):
    parser = subparsers.add_parser(
        'explevel',
        help='check the average expression level of query genes in single cell matrix',
        description=(
            'Check the average expression level of query genes in single cell matrix'
            'Input: processed transcriptome matrix from the process output,'
            'query genes list has to be txt file, genes are seperated with new line.'))

    parser.add_argument('-t', 
                        '--transcriptome_dir', 
                        dest='transcriptome_dir', 
                        required=True,
                        type=str,
                        help='specify the directory from process function.')

    parser.add_argument('-g', 
                        '--gene',
                        dest='gene', 
                        required=True,
                        type=str,
                        help='specify the query genes.')

    parser.add_argument('-o', 
                        '--output_file', 
                        dest='output_file', 
                        required=False,
                        help='specify output file.')

#fc
def add_fc_subparser(subparsers):
    parser = subparsers.add_parser(
        'fc',
        help='check the fold change of genes in given perturbation',
        description=(
            'Check the fold change of perturbed region and individual sgRNA for query region and gene.'
            'Input: processed transcriptome matrix and sgrna matrix from the process output,'
            'sgrna dict file: perturbation region hg38 coordinates and the sgrna sequence targeting that region.'
            'Region and sgrnas separated by tab, and sgrnas separated by comma'))

    parser.add_argument('-t', 
                        '--transcriptome_dir', 
                        dest='transcriptome_dir', 
                        required=True,
                        type=str,
                        help='specify the processed transcriptome matrix file')

    parser.add_argument('-d', 
                        '-dict', 
                        dest='dict', 
                        required=True,
                        type=str, 
                        help='specify the sgRNA annotation file: perturbation coordinates (hg38) and the sgRNA name.')

    parser.add_argument('-r', 
                        '--region', 
                        dest='region', 
                        required=True,
                        type=str,
                        help='specify the query regions and their target genes to calculate repression efficiency.')

    parser.add_argument('-o', 
                        '--output_folder', 
                        dest='output_folder', 
                        required=True,
                        help='specify output folder directory.')

#DEobs
def add_DEobs_subparser(subparsers):
    parser = subparsers.add_parser(
        'DEobs',
        help='perform differential expression analysis of observe cells',
        description=(
            'Perfrom the genome wide differential expression analysis of all the perturbation regions.'
            'Input: processed transcriptome matrix and sgrna matrix from the process output'
            'sgrna dict file: perturbation region hg38 coordinates and the sgrna sequence targeting that region.'
            'region and sgrnas separated by tab, and  sgrnas separated by comma'
            'Output: up regulation p-value, downregulation p-value, fold change(compare with all the otehr cells) and average cpm'))

    parser.add_argument('-t', 
                        '--transcriptome_df', 
                        dest='transcriptome_df', 
                        required=True,
                        type=str,
                        help='specify the processed transcriptome matrix file (.h5)')
    
    parser.add_argument('-s', 
                        '--sgrna', 
                        dest='input_sgrna', 
                        required=True,
                        type=str,
                        help='specify the processed sgrna matrix file. (.h5)')
    
    parser.add_argument('-d', 
                        '-dict', 
                        dest='dict', 
                        required=True,
                        type=str, 
                        help='specify the perturbation coordinates (hg38) and the sgRNA txt file.')

    parser.add_argument('-r', 
                        '--threads', 
                        dest = 'threads', 
                        required=False,
                        type=int,
                        default=1,
                        help='set number of barcode comparison threads. The default is 1')

    parser.add_argument('-n', 
                        '--norm', 
                        dest = 'norm_method', 
                        required=False,
                        type=str,
                        default='cpm',
                        help='choose normalization methods: "cpm" or "metacell".')

    parser.add_argument('-o', 
                        '--output_dir', 
                        dest='output_dir', 
                        required=True,
                        help='specify an output directory.')
#DEobsboot
def add_DEobsboot_subparser(subparsers):
    parser = subparsers.add_parser(
        'DEobsboot',
        help='perform bootstrapping analysis of observe cells',
        description=(
            'Perfrom the genome wide differential expression analysis of all the perturbation regions with bootstrapping strategy.'
            'Input: processed transcriptome matrix and sgrna matrix from the process output'
            'sgrna dict file: perturbation region hg38 coordinates and the sgrna sequence targeting that region.'
            'region and sgrnas separated by tab, and  sgrnas separated by comma'
            'Output: up regulation p-value, downregulation p-value, fold change(compare with all the otehr cells) and average cpm'))

    parser.add_argument('-t', 
                        '--transcriptome_df', 
                        dest='transcriptome_df', 
                        required=True,
                        type=str,
                        help='specify the processed transcriptome matrix file')
    
    parser.add_argument('-s', 
                        '--sgrna', 
                        dest='input_sgrna', 
                        required=True,
                        type=str,
                        help='specify the processed sgrna matrix file.')
    
    parser.add_argument('-d', 
                        '-dict', 
                        dest='dict', 
                        required=True,
                        type=str, 
                        help='specify the perturbation coordinates (hg38) and the sgRNA txt file.')

    parser.add_argument('-r', 
                        '--threads', 
                        dest = 'threads', 
                        required=False,
                        type=int,
                        default=1,
                        help='set number of barcode comparison threads. The default is 1')

    parser.add_argument('-n', 
                        '--norm', 
                        dest = 'norm_method', 
                        required=False,
                        type=str,
                        default='cpm',
                        help='choose normalization methods: "cpm" or "metacell".')

    parser.add_argument('-o', 
                        '--output_dir', 
                        dest='output_dir', 
                        required=True,
                        help='specify an output directory.')
    
#DErand
def add_DErand_subparser(subparsers):
    parser = subparsers.add_parser(
        'DErand',
        help='perform differential expression analysis of random selection background',
        description=(
            'Perfrom the genome wide differential expression analysis of 1000 random selection cells.'
            'There are two options for random selection: all cells with equal probability or probability based on sgrna number in the cells'
            'User should specify the cell number to select randomly.'
            'It is recommended with either exact cell number or bins (large amount of perturbation region).'
            ))

    parser.add_argument('-t', 
                        '--transcriptome_df', 
                        dest='transcriptome_df', 
                        required=True,
                        type=str,
                        help='specify the processed transcriptome matrix file')

    parser.add_argument('-s', 
                        '--sgrna', 
                        dest='input_sgrna', 
                        required=True,
                        type=str,
                        help='specify the processed sgrna matrix file.')

    parser.add_argument('-m', 
                        '--num',
                        dest='num',
                        required=True,
                        type=int,
                        help='specify the number of cells to do random iteration.')

    parser.add_argument('-i',
                        '--iteration',
                        dest='iteration',
                        required=False,
                        type=int,
                        default=1000,
                        help='specify the number of iteration to perform.')

    parser.add_argument('-r', 
                        '--threads', 
                        dest = 'threads', 
                        required=False,
                        type=int,
                        default=1,
                        help='set number of barcode comparison threads. The default is 1')

    parser.add_argument('-n', 
                        '--norm', 
                        dest = 'norm_method', 
                        required=False,
                        type=str,
                        default='cpm',
                        help='choose normalization methods: "cpm" or "metacell".')

    parser.add_argument('-d', 
                        '--randomization_method', 
                        dest = 'randomization_method', 
                        required=True,
                        type=str,
                        help='choose randomization methods: "equal" or "sgrna".')

    parser.add_argument('-o', 
                        '--output_dir', 
                        dest='output_dir', 
                        required=True,
                        help='specify an output directory.')

#local
def add_local_subparser(subparsers):
    parser = subparsers.add_parser(
        'local',
        help='perform local hit analysis (+-2Mb) with observation data and random background',
        description=('Using the observation p value and randomization bavckground p value'
                     'to calculate the adjusted p value based on Gaussian distribution approximation'
                     'Local hits calculation includes the genes within plus and minus 2 Mb of the perturbation region.'
                     'The output is a csv file with all hits information.'))
   
    parser.add_argument('-f', 
                        '--file_dir', 
                        dest='file_dir', 
                        required=True,
                        type=str,
                        help='specify the file directory of "process" function output, the Trans_genome_seq.npy file is required at this step.')
   
    parser.add_argument('-d', 
                        '--data_dir', 
                        dest='data_dir', 
                        required=True,
                        type=str,
                        help='specify the p-value matrix directory of observation test.')

    parser.add_argument('-t', 
                        '--distr', 
                        dest='distr', 
                        required=True,
                        type=str,
                        help='specify the random cell mean/std/10_perc file directory.')
    
    parser.add_argument('-s', 
                        '-sgrna_dict', 
                        dest='sgrna_dict', 
                        required=True,
                        type=str, 
                        help='specify the perturbation coordinates (hg38) and the sgRNA txt file.')

    parser.add_argument('-o', 
                        '--output_file', 
                        dest='output_file', 
                        required=True,
                        type=str,
                        help='specify an output file name incluseing the directory, it has to be in csv format.')

#global
def add_global_subparser(subparsers):
    parser = subparsers.add_parser(
        'global',
        help='perform global hit analysis with observation data and random background',
        description=('Using the observation p value and randomization bavckground p value'
                     'to calculate the adjusted p value based on Gaussian distribution approximation'
                     'The output is a csv file with all hits information.'
                     'cutoff for p value is -1 (ln), fold change is 10%'))
    
    parser.add_argument('-f', 
                        '--file_dir', 
                        dest='file_dir', 
                        required=True,
                        type=str,
                        help='specify the file directory of "process" function output, the Trans_genome_seq.npy file is required at this step.')
    
    parser.add_argument('-d', 
                        '--data_dir', 
                        dest='data_dir', 
                        required=True,
                        type=str,
                        help='specify the p-value matrix directory of observation test. (DEobs output folder)')

    parser.add_argument('-s', 
                        '-sgrna_dict', 
                        dest='sgrna_dict', 
                        required=True,
                        type=str, 
                        help='specify the perturbation coordinates (hg38) and the sgRNA txt file.')
        
    parser.add_argument('-t', 
                        '--distr', 
                        dest='distr', 
                        required=True,
                        type=str,
                        help='specify the random cell file directory. (DErand output folder)')

    parser.add_argument('-o', 
                        '--output_file', 
                        dest='output_file', 
                        required=True,
                        type=str,
                        help='specify an output file name including the directory, it has to be in csv format.')


#gene_expr_level.py

import os
import re
import sys
import collections
import argparse
import tables
import itertools
import matplotlib
import numba
import glob
import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.sparse as sp_sparse

from collections import defaultdict
from scipy import sparse, io
from scipy.sparse import csr_matrix

from importlib import resources
from pySpade.utils import get_logger

logger = get_logger(logger_name=__name__)

def Calculate_gene_expression(TRANSCRIPTOME_DIR, 
                              GENE_FILE, 
                              OUTPUT_FILE):

    TRANSCRIPTOME_FILE = glob.glob(TRANSCRIPTOME_DIR + 'Singlet_sub_df.*')
    if TRANSCRIPTOME_FILE[0].endswith('h5') == True:
        sub_df = pd.read_hdf(TRANSCRIPTOME_FILE[0], 'df')
    if TRANSCRIPTOME_FILE[0].endswith('pkl') == True:
        sub_df = pd.read_pickle(TRANSCRIPTOME_FILE[0])

    #load quired genes
    GENE_LIST = []
    if GENE_FILE.endswith('.txt'):
        with open(GENE_FILE) as ft:
            for line in ft:
                gene = line.strip()
                GENE_LIST.append(gene)
    else:
        for i in GENE_FILE.split(','):
            GENE_LIST.append(i)

    total_seq_reads = sub_df.sum(axis=0)
    logger.info('Finished loading data.')

    #calculate average cpm and write to output file
    if OUTPUT_FILE != None:
        output_file=open(OUTPUT_FILE, 'w')
        output_file.write('Gene' + '\t' + 'Average expression (cpm)' + '\t' + 'Median cpm' + '\t' + 'Portion of cell express' + '\n')
        for gene in GENE_LIST:
            if (gene in sub_df.index) == False:
                logger.critical(str(gene) + ' is missing from the transcriptome dataframe.')
                continue
            dist = sub_df.loc[gene ,:] / total_seq_reads * 1000000
            ave_cpm = np.mean(dist)
            median_cpm = np.median(dist)
            perc_cell = np.sum(dist > 0) /len(dist)
            output_file.write(str(gene) + '\t' + str(ave_cpm) + '\t' + str(median_cpm) + '\t' + str(perc_cell) + '\n')
            logger.info('Finish processing ' + str(gene))
        output_file.close()

    else:
        logger.info('Gene' + '\t' + 'Average expression (cpm)' + '\t' + 'Median cpm' + '\t' + 'Portion of cell express' + '\n')
        for gene in GENE_LIST:
            if (gene in sub_df.index) == False:
                logger.critical(str(gene) + ' is missing from the transcriptome dataframe.')
                continue
            dist = sub_df.loc[gene ,:] / total_seq_reads * 1000000
            ave_cpm = np.mean(dist)
            median_cpm = np.median(dist)
            perc_cell = np.sum(dist > 0) /len(dist)
            logger.info(str(gene) + '\t' + str(ave_cpm) + '\t' + str(median_cpm) + '\t' + str(perc_cell) + '\n')

    

if __name__ == '__main__':
    pass
